@@ -2,8 +2,8 @@
 
 about = {
     "name": "softchat",
-    "version": "1.2",
-    "date": "2021-04-01",
+    "version": "1.5",
+    "date": "2023-02-15",
     "description": "convert twitch/youtube chat into softsubs",
     "author": "ed",
     "license": "MIT",
@@ -27,7 +27,6 @@ import shutil
 import argparse
 import tempfile
 import colorsys
-import queue
 import multiprocessing
 import subprocess as sp
 from PIL import Image
@@ -36,6 +35,7 @@ from .util import WINDOWS, HAVE_FONTFORGE
 from .util import shell_esc, zopen, tt, hms, get_ff_dur, load_fugashi
 from .mproc import TextStuff, gen_msg_thr, gen_msg_initializer
 from .ass import assan, segment_msg, render_msegs
+from .fconv import convert_file
 
 
 try:
@@ -107,12 +107,30 @@ def cache_emotes(emotes, emote_dir, overwrite):
 
         if not source_ok:
             url = None
+            alts = []
+            best = 0
             for img in e["images"]:
                 if img["id"] == "source":
                     url = img["url"]
+                    break
+
+                w = img["width"]
+                if w < best:
+                    continue
+
+                if w > best:
+                    best = w
+                    alts = []
+
+                alts.append(img)
+                if "dark" in img["id"]:
+                    url = img["url"]
+
+            if url is None and alts:
+                url = alts[-1]["url"]
 
             if url is None:
-                error(f"Could not find source URL for {e['name']}")
+                error(f"Could not find URL for {e['name']}")
                 sys.exit(1)
 
             r = requests.get(url)
@@ -256,9 +274,11 @@ def main():
     ap.add_argument("--kana", action="store_true", help="convert kanji to kana")
     ap.add_argument("--fontdir", metavar="DIR", type=str, default=None, help="path to noto-hinted")
     ap.add_argument("--dupe_thr", metavar="SEC", type=float, default=10, help="Hide duplicate messages from the same author within this many seconds")
+    ap.add_argument("--filter_gifts", action="store_true", help="Filter out the repetitive 'was gifted a membership by X' messages from gifted memberships")
     ap.add_argument("--no_del", action="store_true", help="keep msgs deleted by mods")
     ap.add_argument("--start_time", metavar="STRT", type=int, default=None, help="Start time of the video in as a unix timestamp in seconds. Only used when there is no VOD chat download.")
     ap.add_argument("--offset", metavar="OFS", type=float, default=None, help="Offset in seconds to apply to the chat. Positive values delay the chat, negative values advance the chat, the same as subtitle delay in MPV. Use with incomplete video downloads or when estimating the start time.")
+    ap.add_argument("--badge_sz", metavar="MUL", type=float, default=2, help="Multiplier for VIP icon size")
     ap.add_argument("--emote_font", action="store_true", help="Generate a custom emote font for emotes found in this stream. The subtitle file produced will require the specific embedded font generated by this run to be embedded in the media file.")
     ap.add_argument("--emote_cache", metavar="EMOTE_DIR", type=str, default=None, help="Directory to store emotes in. By default it is $pwd/emotes, but using the same directory for all invocations is safe. Will be created if it does not exist.")
     ap.add_argument("--emote_sz", metavar="MUL", type=float, default=1, help="Emote size multiplier")
@@ -302,10 +322,10 @@ def main():
     if ar.media and os.path.isfile(ar.media):
         media_fn = ar.media
     else:
-        for ext in ["webm", "mp4", "mkv"]:
-            f = ar.fn[0]
-            while not media_fn and "." in f:
-                f = f.rsplit(".", 1)[0]
+        f = ar.fn[0]
+        while not media_fn and "." in f:
+            f = f.rsplit(".", 1)[0]
+            for ext in ["webm", "mp4", "mkv"]:
                 mfn = f + "." + ext
                 if os.path.isfile(mfn):
                     media_fn = mfn
@@ -316,7 +336,7 @@ def main():
                     media_fn = mfn
                     break
 
-    base_fn = ar.fn[0].rsplit(".json", 1)[0]
+    base_fn = ar.fn[0].rsplit(".json", 1)[0].rsplit(".live_chat", 1)[0]
     if media_fn:
         base_fn = media_fn.rsplit(".", 1)[0]
 
@@ -364,6 +384,18 @@ def main():
             except Exception as ex:
                 err = repr(ex)
 
+            if err and "Extra data" in err:
+                m = "input json is not a valid chat_downloader chatlog; trying to convert it..."
+                warn(m)
+                try:
+                    jd2 = list(convert_file(f))
+                    # debug("writing converted json")
+                    # with open(fn + ".conv.json", "w", encoding="utf-8") as cf:
+                    #     json.dump(jd2, cf, sort_keys=True, indent=2)
+                    err = None
+                except Exception as ex:
+                    err += "\nfconv: " + repr(ex)
+
             if not err and not jd2:
                 err = "empty json file?"
 
@@ -376,7 +408,7 @@ def main():
 
             if not err:
                 try:
-                    _ = jd2[0]["timestamp"]
+                    _ = jd2[0]["action_type"]
                 except:
                     err = "does not look like a chatlog"
 
@@ -395,6 +427,54 @@ def main():
                 elif at == "mark_chat_items_by_author_as_deleted":
                     deleted_authors.add(m["author"]["id"])
 
+                if (
+                    ar.filter_gifts
+                    and m.get("message_type", None) == "sponsorships_gift_redemption_announcement"
+                ):
+                    continue
+
+
+                if "emotes" in m:
+                    customs = []
+                    stocks = []
+                    for x in m["emotes"]:
+                        if x.get("is_custom_emoji", True):
+                            customs.append(x)
+                        else:
+                            stocks.append(x)
+
+                    # only keep/convert the custom emotes
+                    m["emotes"] = customs
+
+                    # non-customs have regular unicode emojis as IDs,
+                    # so swap out the shortcuts with those instead
+                    for emote in stocks:
+                        uchar = emote["id"]
+                        if len(uchar) > 8:
+                            continue
+
+                        txt = m["message"]
+                        if emote["shortcuts"] is not None:
+                            for sc in emote["shortcuts"]:
+                                txt = txt.replace(sc, uchar)
+
+                        m["message"] = txt
+
+                if at is None and "message" in m:
+                    # twitch
+                    msg = m["message"]
+                    at = "add_chat_item"
+                    for emote in m.get("emotes", []):
+                        if "shortcuts" in emote:
+                            warn(f"expected no shortcuts, got [{emote['shortcuts']}]")
+                            continue
+
+                        shortcut = ":" + emote["name"] + ":"
+                        msg = msg.replace(emote["name"], shortcut)
+                        emote["shortcuts"] = [shortcut]
+
+                    m["message"] = msg
+
                 if (
                     at != "add_chat_item"
                     or m.get("author", {}).get("id", None) is None
@@ -494,6 +574,29 @@ def main():
         unix_ofs = ar.start_time
 
     if unix_ofs is None and ar.offset is None:
+        # build a rough histogram of the first 5 minutes,
+        # the first burst of hype is probably close
+        info("cannot continue without --start_time; here are some hints:")
+        htime = -1
+        hcount = 0
+        hints = 0
+        htxt = ""
+        for x in jd:
+            t = int(x["timestamp"] / 1_000_000)
+            if htime == t:
+                hcount += 1
+                continue
+
+            if hcount:
+                info(f'{hcount} messages at start_time {htime} "{htxt}"')
+
+            htime = t
+            hcount = 1
+            htxt = x["message"]
+            hints += 1
+            if hints > 300:
+                break
+
         raise Exception(
             "could not find time_in_seconds in json, set a start time or offset "
             "manually with --start_time/--offset or use v0.17 or earlier"
@@ -678,9 +781,8 @@ def main():
     for n_msg, msg, vtxt, vsz, t_fsec, t_hms, msg_emotes in gen_msgs(
         jd, vw, bw, ar, emote_shortcuts, have_fugashi
     ):
-        # pillow height calculation is off by a bit; this is roughly it i think
         sx, sy = vsz
-        sy = int(sy + fofs * 2.5 + 0.8)
+        sy = int(sy - 10)
 
         if n_msg % 1000 == 0:
             info(
@@ -821,32 +923,45 @@ Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
 
             if ar.m == 1:
                 # text = colored nick followed by the actual lines, ass-escaped
-                txt = [rf"{{\3c&H{bgr_nick}&}}{nick}"]
+                txt = [rf"{{\3c&H{bgr_nick}&\fs{ar.sz*0.67:.1f}}}{nick}"]
+                lineh = z.emote_vsz[1]
+                nickh = lineh * 0.7
 
                 badges = msg.get("badges", [])
                 if not set(badges).isdisjoint(["Moderator", "Owner", "Verified"]):
-                    txt[-1] += r" {\bord16\shad6}*"
+                    txt[0] += r" {\bord16\shad6}*"
                 elif msg["uid"] in vips:
-                    txt[-1] += r" {\bord16\shad4}----"
+                    txt[0] += r" {\bord16\shad4}----"
 
                 if shrimp:
-                    txt.append(shrimp)
+                    txt.append(rf"{{\fscx90\fscy90}}{shrimp}{{\fscx100\fscy100}}")
+                    msg["sy"] += lineh
 
                 txt.extend(msg["txt"])
+                txt[1] = rf"{{\fs{ar.sz}}}{txt[1]}"
+
+                if msg["msg_emotes"] and (
+                    ar.emote_fill or filled_emotes or ar.emote_sz > 1.01
+                ):
+                    txt = "\n".join(txt)
+                    msegs = segment_msg(txt, ar.emote_fill, filled_emotes)
+                    txt = render_msegs(
+                        msegs, ar.sz, ar.sz * ar.emote_sz, bgr_msg, bgr_fg, bord, shad
+                    )
+                    txt = txt.split("\n")
 
                 # show new messages bottom-left (heh)
                 msg["txt"] = txt
                 msg["px"] = bx
-                msg["py"] = by + bh
+                msg["py"] = by + bh - msg["sy"] - nickh
 
                 ta = hms(msg["t0"])
                 tb = hms(next_msg["t0"] if next_msg else msg["t0"] + 10)
 
                 rm = 0
                 for m in vis:
-                    # i'll be honest, no idea why the *1.2 is necessary
-                    m["py"] -= msg["sy"] * 1.2
-                    if m["py"] - m["sy"] < by:
+                    m["py"] -= msg["sy"] + nickh
+                    if m["py"] < by:
                         # debug('drop {} at {}'.format(hms(m["t0"]), ta))
                         rm += 1
 
@@ -854,41 +969,39 @@ Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
 
                 if True:
                     # rely on squished font for linespacing reduction
-                    txt = r"{{\pos({},{})}}".format(bx, by + bh)
+                    txt = r"{{\pos({:.1f},{:.1f})}}".format(bx, by + bh)
                     for m in vis:
                         pad = ""
                         for ln in m["txt"]:
                             txt += pad + ln + r"\N{\r}"
                             pad = r"\h\h"
 
-                    f.write(
-                        "Dialogue: 0,{},{},a,,0,0,0,,{}\n".format(ta, tb, txt).encode(
-                            "utf-8"
-                        )
-                    )
+                    f.write(f"Dialogue: 0,{ta},{tb},a,,0,0,0,,{txt}\n".encode("utf-8"))
                 else:
-                    # old approach, one ass entry for each line
+                    # verification, one ass entry for each line
                     for m in vis:
-                        step = m["sy"] / len(m["txt"])
+                        step = nickh
                         x = m["px"]
                         y = m["py"]
                         for ln in m["txt"]:
-                            txt = r"{{\pos({},{})}}{}".format(x, y, ln)
+                            txt = r"{{\pos({:.1f},{:.1f})}}{}".format(x, y, ln)
                             x = m["px"] + 8
                             y += step
-                            f.write(
-                                "Dialogue: 0,{},{},a,,0,0,0,,{}\n".format(
-                                    ta, tb, txt
-                                ).encode("utf-8")
-                            )
+                            step = m["sy"] / (len(m["txt"]) - 1)
+                            txt = f"Dialogue: 0,{ta},{tb},a,,0,0,0,,{txt}\n"
+                            f.write(txt.encode("utf-8"))
+
             else:
                 txts, t0, w, h, msg_emotes = [
                     msg[x] for x in ["txt", "t0", "sx", "sy", "msg_emotes"]
                 ]
                 txt = "\\N".join(txts)
 
+                lineh = z.emote_vsz[1]
+                nickh = lineh * 0.6
+
                 # ass linespacing is huge, compensate (wild guess btw)
-                h += int(ar.sz * 0.25 * (len(txts) - 1) + 0.99)
+                h += int(ar.sz * 0.25 * (len(txts) - 1) + 0.99 + nickh)
 
                 # plus some horizontal margin between the messages
                 w += 8
@@ -899,12 +1012,18 @@ Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
                     shrimp_mul = 2
                     w += h * 4  # donation is not included, TODO maybe
 
-                if "badges" in msg and "Moderator" in msg["badges"]:
+                vip = False
+                badges = msg.get("badges", [])
+                if not set(badges).isdisjoint(["Moderator", "Owner", "Verified"]):
                     bord = shad = 6
-                    txt = rf"{{\bord24\shad{shad}}}*{{\bord{bord}}}{txt}"
+                    badge_sz = int(12 * ar.badge_sz)
+                    txt = rf"{{\bord{badge_sz}\shad{shad}}}*{{\bord{bord}}}{txt}"
+                    vip = True
                 elif msg["uid"] in vips:
                     bord = shad = 4
-                    txt = rf"{{\bord16\shad{shad}}}_{{\bord{bord}}}{txt}"
+                    badge_sz = int(8 * ar.badge_sz)
+                    txt = rf"{{\bord{badge_sz}\shad{shad}}}_{{\bord{bord}}}{txt}"
+                    vip = True
 
                 len_boost = 0.7
                 td = (vw + w - w * len_boost) * shrimp_mul / ar.spd
@@ -955,7 +1074,7 @@ Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
                 for m in rm:
                     vis.remove(m)
 
-                ymax = vh - h
+                ymax = bh - h
                 if ymax < 1:
                     ymax = 1  # thx emotes
 
@@ -1030,7 +1149,7 @@ Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
                 else:
                     y = None
                     best = 5318008
-                    target = vh / 2
+                    target = bh / 2
                     # print(json.dumps([target, frees], indent=4, sort_keys=True))
                     for avail, y0, y1 in frees:
                         if y0 <= target and y1 >= target + h:
@@ -1056,7 +1175,7 @@ Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
                     )
 
                 y = int(y)
-                txt = rf"{{\move({vw},{y+h},{-w},{y+h})\3c&H{bgr_nick}&}}{txt}{{\fscx40\fscy40\bord1}}\N{nick}"
+                txt = rf"{{\move({vw:.1f},{y+h+by:.1f},{-w:.1f},{y+h+by:.1f})\3c&H{bgr_nick}&}}{txt}{{\fscx40\fscy40\bord1}}\N{nick}"
 
                 msg["t1"] = t1
                 msg["p9"] = p9
@@ -1074,7 +1193,7 @@ Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
                     hms(t0), hms(msg["t1"]), msg["txt"]
                 ).encode("utf-8")
 
-                if shrimp:
+                if shrimp or vip:
                     supers.append(ln)
                 else:
                     f.write(ln)
@@ -1195,13 +1314,13 @@ r"""
 
 ===[ side-by-side with kanji transcription ]===========================
 
-chat_replay_downloader.py https://www.youtube.com/watch?v=7LXgVlrfsWw -output ars-37-minecraft-3.json
+ytdl-tui.py https://www.youtube.com/watch?v=7LXgVlrfsWw
 
-softchat.py ..\yt\ars-37-minecraft-3.json -b 340x500+32+32 && copy /y ..\yt\ars-37-minecraft-3.ass ..\yt\ars-37-minecraft-3.1.ass
+softchat.py ..\yt\ars-37-minecraft-3.json -b 360x500+24+40 --sz 26 --emote_font --emote_sz 1.8 && copy /y ..\yt\ars-37-minecraft-3.ass ..\yt\ars-37-minecraft-3.1.ass
 
-softchat.py ..\yt\ars-37-minecraft-3.json -b 340x500+360+32 --kana && copy /y ..\yt\ars-37-minecraft-3.ass ..\yt\ars-37-minecraft-3.2.ass
+softchat.py ..\yt\ars-37-minecraft-3.json -b 360x500+300+40 --sz 26 --emote_font --emote_sz 1.8 --kana && copy /y ..\yt\ars-37-minecraft-3.ass ..\yt\ars-37-minecraft-3.2.ass
 
-(head -n 21 ars-37-minecraft-3.1.ass ; (tail -n +22 ars-37-minecraft-3.1.ass; tail -n +22 ars-37-minecraft-3.2.ass) | sort) > ars-37-minecraft-3.ass
+(head -n 22 ars-37-minecraft-3.1.ass ; (tail -n +23 ars-37-minecraft-3.1.ass; tail -n +23 ars-37-minecraft-3.2.ass) | sort) > ars-37-minecraft-3.ass
 
 c:\users\ed\bin\mpv.com "..\yt\ars-37-minecraft-3.json.mkv" -ss 1:15:20
 
